@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2001-2008 Marc Alexander Lehmann <schmorp@schmorp.de>
+ * Copyright (c) 2001-2009 Marc Alexander Lehmann <schmorp@schmorp.de>
  * 
  * Redistribution and use in source and binary forms, with or without modifica-
  * tion, are permitted provided that the following conditions are met:
@@ -117,6 +117,8 @@ trampoline (int sig)
        ".globl coro_transfer\n"
        ".type coro_transfer, @function\n"
        "coro_transfer:\n"
+       /* windows, of course, gives a shit on the amd64 ABI and uses different registers */
+       /* http://blogs.msdn.com/freik/archive/2005/03/17/398200.aspx */
        #if __amd64
          #define NUM_SAVED 6
          "\tpush %rbp\n"
@@ -125,8 +127,18 @@ trampoline (int sig)
          "\tpush %r13\n"
          "\tpush %r14\n"
          "\tpush %r15\n"
+         #if CORO_WIN_TIB
+           "\tpush %gs:0x0\n"
+           "\tpush %gs:0x8\n"
+           "\tpush %gs:0xc\n"
+         #endif
          "\tmov  %rsp, (%rdi)\n"
          "\tmov  (%rsi), %rsp\n"
+         #if CORO_WIN_TIB
+           "\tpop  %gs:0xc\n"
+           "\tpop  %gs:0x8\n"
+           "\tpop  %gs:0x0\n"
+         #endif
          "\tpop  %r15\n"
          "\tpop  %r14\n"
          "\tpop  %r13\n"
@@ -139,8 +151,18 @@ trampoline (int sig)
          "\tpush %ebx\n"
          "\tpush %esi\n"
          "\tpush %edi\n"
+         #if CORO_WIN_TIB
+           "\tpush %fs:0\n"
+           "\tpush %fs:4\n"
+           "\tpush %fs:8\n"
+         #endif
          "\tmov  %esp, (%eax)\n"
          "\tmov  (%edx), %esp\n"
+         #if CORO_WIN_TIB
+           "\tpop  %fs:8\n"
+           "\tpop  %fs:4\n"
+           "\tpop  %fs:0\n"
+         #endif
          "\tpop  %edi\n"
          "\tpop  %esi\n"
          "\tpop  %ebx\n"
@@ -190,8 +212,8 @@ coro_create (coro_context *ctx, coro_func coro, void *arg, void *sptr, long ssiz
     }
 
   /* set the new stack */
-  nstk.ss_sp    = STACK_ADJUST_PTR (sptr,ssize); /* yes, some platforms (IRIX) get this wrong. */
-  nstk.ss_size  = STACK_ADJUST_SIZE (sptr,ssize);
+  nstk.ss_sp    = STACK_ADJUST_PTR (sptr, ssize); /* yes, some platforms (IRIX) get this wrong. */
+  nstk.ss_size  = STACK_ADJUST_SIZE (sptr, ssize);
   nstk.ss_flags = 0;
 
   if (sigaltstack (&nstk, &ostk) < 0)
@@ -225,21 +247,24 @@ coro_create (coro_context *ctx, coro_func coro, void *arg, void *sptr, long ssiz
 # elif CORO_LOSER
 
   coro_setjmp (ctx->env);
-  #if __CYGWIN__
-    ctx->env[7] = (long)((char *)sptr + ssize) - sizeof (long);
-    ctx->env[8] = (long)coro_init;
+  #if __CYGWIN__ && __i386
+    ctx->env[8]                        = (long)    coro_init;
+    ctx->env[7]                        = (long)    ((char *)sptr + ssize)         - sizeof (long);
+  #elif __CYGWIN__ && __x86_64
+    ctx->env[7]                        = (long)    coro_init;
+    ctx->env[6]                        = (long)    ((char *)sptr + ssize)         - sizeof (long);
   #elif defined(__MINGW32__)
-    ctx->env[4] = (long)((char *)sptr + ssize) - sizeof (long);
-    ctx->env[5] = (long)coro_init;
+    ctx->env[5]                        = (long)    coro_init;
+    ctx->env[4]                        = (long)    ((char *)sptr + ssize)         - sizeof (long);
   #elif defined(_M_IX86)
-    ((_JUMP_BUFFER *)&ctx->env)->Eip   = (long)coro_init;
-    ((_JUMP_BUFFER *)&ctx->env)->Esp   = (long)STACK_ADJUST_PTR (sptr, ssize) - sizeof (long);
+    ((_JUMP_BUFFER *)&ctx->env)->Eip   = (long)    coro_init;
+    ((_JUMP_BUFFER *)&ctx->env)->Esp   = (long)    STACK_ADJUST_PTR (sptr, ssize) - sizeof (long);
   #elif defined(_M_AMD64)
-    ((_JUMP_BUFFER *)&ctx->env)->Rip   = (__int64)coro_init;
-    ((_JUMP_BUFFER *)&ctx->env)->Rsp   = (__int64)STACK_ADJUST_PTR (sptr, ssize) - sizeof (long);
+    ((_JUMP_BUFFER *)&ctx->env)->Rip   = (__int64) coro_init;
+    ((_JUMP_BUFFER *)&ctx->env)->Rsp   = (__int64) STACK_ADJUST_PTR (sptr, ssize) - sizeof (__int64);
   #elif defined(_M_IA64)
-    ((_JUMP_BUFFER *)&ctx->env)->StIIP = (__int64)coro_init;
-    ((_JUMP_BUFFER *)&ctx->env)->IntSp = (__int64)STACK_ADJUST_PTR (sptr, ssize) - sizeof (long);
+    ((_JUMP_BUFFER *)&ctx->env)->StIIP = (__int64) coro_init;
+    ((_JUMP_BUFFER *)&ctx->env)->IntSp = (__int64) STACK_ADJUST_PTR (sptr, ssize) - sizeof (__int64);
   #else
     #error "microsoft libc or architecture not supported"
   #endif
@@ -248,17 +273,17 @@ coro_create (coro_context *ctx, coro_func coro, void *arg, void *sptr, long ssiz
 
   coro_setjmp (ctx->env);
   #if __GLIBC__ >= 2 && __GLIBC_MINOR__ >= 0 && defined (JB_PC) && defined (JB_SP)
-    ctx->env[0].__jmpbuf[JB_PC] = (long)coro_init;
-    ctx->env[0].__jmpbuf[JB_SP] = (long)STACK_ADJUST_PTR (sptr, ssize) - sizeof (long);
+    ctx->env[0].__jmpbuf[JB_PC]        = (long)    coro_init;
+    ctx->env[0].__jmpbuf[JB_SP]        = (long)    STACK_ADJUST_PTR (sptr, ssize) - sizeof (long);
   #elif __GLIBC__ >= 2 && __GLIBC_MINOR__ >= 0 && defined (__mc68000__)
     ctx->env[0].__jmpbuf[0].__aregs[0] = (long int)coro_init;
-    ctx->env[0].__jmpbuf[0].__sp = (int *)((char *)sptr + ssize) - sizeof (long);
+    ctx->env[0].__jmpbuf[0].__sp       = (int *)   ((char *)sptr + ssize)         - sizeof (long);
   #elif defined (__GNU_LIBRARY__) && defined (__i386__)
-    ctx->env[0].__jmpbuf[0].__pc = (char *)coro_init;
-    ctx->env[0].__jmpbuf[0].__sp = (void *)((char *)sptr + ssize) - sizeof (long);
+    ctx->env[0].__jmpbuf[0].__pc       = (char *)  coro_init;
+    ctx->env[0].__jmpbuf[0].__sp       = (void *)  ((char *)sptr + ssize)         - sizeof (long);
   #elif defined (__GNU_LIBRARY__) && defined (__amd64__)
-    ctx->env[0].__jmpbuf[JB_PC]  = (long)coro_init;
-    ctx->env[0].__jmpbuf[0].__sp = (void *)((char *)sptr + ssize) - sizeof (long);
+    ctx->env[0].__jmpbuf[JB_PC]        = (long)    coro_init;
+    ctx->env[0].__jmpbuf[0].__sp       = (void *)  ((char *)sptr + ssize)         - sizeof (long);
   #else
     #error "linux libc or architecture not supported"
   #endif
@@ -266,14 +291,21 @@ coro_create (coro_context *ctx, coro_func coro, void *arg, void *sptr, long ssiz
 # elif CORO_IRIX
 
   coro_setjmp (ctx->env, 0);
-  ctx->env[JB_PC] = (__uint64_t)coro_init;
-  ctx->env[JB_SP] = (__uint64_t)STACK_ADJUST_PTR (sptr, ssize) - sizeof (long);
+  ctx->env[JB_PC]                      = (__uint64_t)coro_init;
+  ctx->env[JB_SP]                      = (__uint64_t)STACK_ADJUST_PTR (sptr, ssize) - sizeof (long);
 
 # elif CORO_ASM
 
   ctx->sp = (void **)(ssize + (char *)sptr);
   *--ctx->sp = (void *)abort; /* needed for alignment only */
   *--ctx->sp = (void *)coro_init;
+
+  #if CORO_WIN_TIB
+  *--ctx->sp = 0;                    /* ExceptionList */
+  *--ctx->sp = (char *)sptr + ssize; /* StackBase */
+  *--ctx->sp = sptr;                 /* StackLimit */
+  #endif
+
   ctx->sp -= NUM_SAVED;
 
 # elif CORO_UCONTEXT
@@ -372,7 +404,12 @@ coro_create (coro_context *ctx, coro_func coro, void *arg, void *sptr, long ssiz
       args.main = &nctx;
 
       pthread_attr_init (&attr);
+#if __UCLIBC__
+      /* exists, but is borked */
+      /*pthread_attr_setstacksize (&attr, (size_t)ssize);*/
+#else
       pthread_attr_setstack (&attr, sptr, (size_t)ssize);
+#endif
       pthread_attr_setscope (&attr, PTHREAD_SCOPE_PROCESS);
       pthread_create (&ctx->id, &attr, coro_init, &args);
 
